# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/8/31 下午3:39
#  @brief: 
# ===============================================================

from .dao import Dao
from ..common import constants, util

sqls = constants.Sql
tb_names = constants.TBName


class TaskDao(Dao):
    def __init__(self, redis_config, mysql_config):
        # 强制设置监控db
        redis_config = {
            **redis_config, **{'db': constants.REDIS_DB}
        }
        super().__init__(redis_config, mysql_config)

    def store_monitor_info(self, key):
        """存储监控信息"""
        value = self.redis_client.hgetall(key)
        value = {k.decode(): v.decode() for k, v in value.items()}
        value['task_key'] = key
        # 计算平均爬取时间
        value = util.get_avg_crawl_time(value)
        sql = sqls.INSERT + f'on duplicate key update success=values(success), ' \
                            f'failed=values(failed), retry=values(retry), avg_crawl_time=values(avg_crawl_time)'
        self.store(sql, tb_names.TASK_MONITOR, value)

    def sum_by_task(self, date):
        """统计任务维度下的整体统计信息"""
        sql = f'select `bis_name`, `task_name`, sum(`queue_length`) `queue_length`, sum(`left`) `left`, ' \
              f'sum(`success`) `success`, sum(`failed`) `failed`, sum(`retry`) `retry` ' \
              f'from `{tb_names.TASK_MONITOR}` where `version` like "{date}%" ' \
              f'group by `bis_name`, `task_name`'
        return self.get(sql)

    def get_tasks(self, date):
        """获取某日的所有任务"""
        sql = f'select distinct `bis_name`, `task_name` from `{tb_names.TASK_MONITOR}` ' \
              f'where `version` like "{date}%"'
        return self.get(sql)

    def get_task_versions(self, bis_name, task_name, date):
        """获取某日某任务的所有批次"""
        sql = f'select `queue_length`, `left`, `success`, `failed`, `retry`, `avg_crawl_time`, `version` ' \
              f'from `{tb_names.TASK_MONITOR}` ' \
              f'where `bis_name` = "{bis_name}" and `task_name` = "{task_name}" and `version` like "{date}%"'
        return self.get(sql)



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/8/31 上午9:24
#  @brief: 报表
# ===============================================================

import logging
from pandas import DataFrame
from datetime import date, timedelta

from ..common import util, constants
from ..core.mail import Mail
from ..core.task_dao import TaskDao

email_address = constants.EmailAddress

test = constants.TEST


class Report:
    def __init__(self, redis_config, mysql_config, mail_config):
        self.dao = TaskDao(redis_config, mysql_config)
        self.redis_client = self.dao.redis_client
        self.mail = Mail(mail_config)
        self.yesterday, self.two_days_ago = None, None
        self.keys = self.match_keys()

    def match_keys(self):
        """获取指定的keys"""
        # 获取前两天的日期
        today = date.today()
        yesterday = today + timedelta(days=-1)
        two_days_ago = today + timedelta(days=-2)
        self.yesterday = yesterday.strftime(constants.DATE_FMT)
        self.two_days_ago = two_days_ago.strftime(constants.DATE_FMT)
        # 匹配key
        keys = self.redis_client.keys(f'*{self.yesterday}*')
        keys.extend(self.redis_client.keys(f'*{self.two_days_ago}*'))
        return keys

    def report(self):
        """发送报表"""
        # 存储监控信息
        for key in self.keys:
            try:
                self.dao.store_monitor_info(key.decode())
            except Exception as e:
                logging.exception(e)
        # 获取昨日的所有任务
        tasks = self.dao.get_tasks(self.yesterday)
        if len(tasks) == 0:
            return
        # 获取昨日任务维度下的整体统计信息
        task_statistic = self.dao.sum_by_task(self.yesterday)
        df = DataFrame(task_statistic)
        df = df[['bis_name', 'task_name', 'queue_length', 'left', 'success', 'failed', 'retry']]
        html = util.df_to_html('整体统计信息', df)
        # 获取昨日各任务的统计信息
        for task in tasks:
            bis_name = task['bis_name']
            task_name = task['task_name']
            task_statistic = self.dao.get_task_versions(bis_name, task_name, self.yesterday)
            df = DataFrame(task_statistic)
            df = df[['queue_length', 'left', 'success', 'failed', 'retry', 'avg_crawl_time', 'version']]
            html += util.df_to_html(f'业务: {bis_name} 任务: {task_name} 统计信息', df)
        # 发送报表
        if test:
            self.mail.send('爬虫每日统计报表', html, addresses=[email_address.HUANGCHI])
        else:
            self.mail.send('爬虫每日统计报表', html)



# -*- coding: utf-8 -*-
# ===============================================================
#  author: hehuihui@zhiyitech.com
#  date: 2018/04/18 14:17
#  brief:
# ===============================================================

import time
import smtplib
import logging
from email.mime.text import MIMEText
from email.utils import formataddr

from ..common import constants

email_addresses = constants.EmailAddress

# 默认的收件人列表
default_addresses = [email_addresses.ZEYU, email_addresses.HUIHUI,
                     email_addresses.XIANGYOU, email_addresses.HUANGCHI]


class Mail:
    def __init__(self, mail_config, nickname='spider', time_interval: int=0):
        """
        :param mail_config: 邮件配置，服务器、端口、用户名及密码
        :param nickname: 发件人昵称
        :param time_interval: 发送时间间隔
        """
        self.mail_config = mail_config
        # 昵称
        self.nickname = nickname
        # 发送时间间隔
        self.time_interval = time_interval
        # 上次发送邮件的时间
        self.last_mail_time = None

    def send(self, subject, content, addresses: list=default_addresses,
             cc_addresses: list=None, subtype='html', charset='utf-8'):
        """
        发送邮件
        :param subject: 邮件标题
        :param content: 邮件内容
        :param addresses: 收件人列表
        :param cc_addresses: 抄送人列表
        :param subtype: 邮件格式（例如：plain/html/json等）
        :param charset: 邮件字符编码
        :return: None
        """
        # 若设置了发送时间间隔
        # 判断距上一封邮件的发送时间是否超过 time_interval
        # 若还在 time_interval 内，不发送邮件
        if self.time_interval:
            if not self.last_mail_time:
                self.last_mail_time = int(time.time())
            elif (time.time() - self.last_mail_time) < self.time_interval:
                left_time = int(time.time() - self.last_mail_time)
                logging.debug(f'mail system not ready, left {left_time} seconds...')
                return
            else:
                self.last_mail_time = int(time.time())
        logging.debug('send mail start...')
        # 构造邮件对象
        mail_config = self.mail_config
        user = mail_config['user']
        msg = MIMEText(content, subtype, charset)
        msg['From'] = formataddr([self.nickname, user])
        if addresses:
            msg['To'] = ','.join(addresses)
        if cc_addresses:
            msg['CC'] = ','.join(cc_addresses)
        msg['Subject'] = subject
        server = smtplib.SMTP_SSL('smtp.exmail.qq.com', 465)
        server.login(user, mail_config['password'])
        server.sendmail(user, addresses, msg.as_string())
        server.quit()
        logging.debug('send mail end...')



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/9/17 上午11:49
#  @brief: 
# ===============================================================

import setuptools

with open("spider_task_monitor/README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spider_task_monitor",
    version="0.0.6",
    author="IceFruit",
    author_email="huangchi@zhiyitech.cn",
    description="对于爬虫任务调度的监控",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IceFruit0777/py_package/tree/master/spider_task_monitor",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-
# ===============================================================
#  @author: huangchi@zhiyitech.cn
#  @date: 2018/9/11 上午8:34
#  @brief: 
# ===============================================================

import csv
import json
import redis
import hashlib
from functools import wraps
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


def task_detect(task, task_config):
    """
    监控任务的检测，对任务是否已完成，任务完成质量的检测
    :param task: 任务信息
    :param task_config: 任务的配置信息
    :return: content，报警信息，若检测通过返回空字符串
    """
    content = ''
    # 任务是否完成的检测
    queue_length = int(task[b'queue_length'])
    left = int(task[b'left'])
    if left < 0:
        # 数据异常
        content += f'Abnormal data:: left: {left}<br />'
    else:
        if left > 0:
            content += f'任务未完成:: {left} left<br />'
        # 失败率检测
        if task_config.get(b'failure_rate_up') and task.get(b'failed'):
            failure_rate = float(task[b'failed']) / (queue_length - left)
            failure_rate_up = float(task_config[b'failure_rate_up'])
            if failure_rate > failure_rate_up:
                content += f'失败率超标:: {round(failure_rate, 2)} > {failure_rate_up}<br />'
        # 爬取时长检测
        if task_config.get(b'avg_time_up') and task.get(b'crawl_time'):
            avg_time = float(task[b'crawl_time']) / (queue_length - left)
            avg_time_up = float(task_config[b'avg_time_up'])
            if avg_time > avg_time_up:
                content += f'爬取时长超标:: {round(avg_time, 2)} > {avg_time_up}'
    return content


def get_avg_crawl_time(value):
    """获取任务监控的平均爬取时间"""
    crawl_time = value.get('crawl_time')
    if not crawl_time:
        return value
    crawl_time = float(crawl_time)
    value.pop('crawl_time')
    # 总爬取次数
    total_crawl_num = int(value.get('success', 0)) + int(value.get('failed', 0)) + int(value.get('retry', 0))
    if total_crawl_num == 0:
        return value
    value['avg_crawl_time'] = crawl_time / total_crawl_num
    return value


def get_dict_hash(data):
    """将字典进行hash"""
    data = json.dumps(data).encode()
    md5_code = hashlib.md5(data).hexdigest()
    return md5_code


def get_redis_conn(redis_config):
    """获取redis连接池"""
    return redis.Redis(connection_pool=redis.ConnectionPool(**redis_config))


def get_mysql_conn(mysql_config, sql_engine_config):
    """获取mysql连接池"""
    uri = 'mysql+pymysql://{user}:{passwd}@{host}:{port}/{db}?charset={charset}'.format(**mysql_config)
    engine = create_engine(uri, **sql_engine_config)
    session = sessionmaker(bind=engine)
    return session


def get_session_fmt(keys):
    """将keys转化为sqlalchemy的session对象可读的预定义格式化字符串"""
    return ','.join(f'`{key}`=:{key}' for key in keys)


def export_as_csv(filename, *data):
    """以csv导出文件"""
    if len(data) == 0:
        return
    writer = csv.DictWriter(open(filename, 'a+'), fieldnames=data[0].keys())
    writer.writeheader()
    writer.writerows(data)


def sql_engine():
    """调用sqlalchemy连接进行mysql交互的装饰器"""
    def decorator(func):
        @wraps(func)
        def wrapper(cls, *args, **kwargs):
            """
            :param cls: 当前类型的实例
            :param args:
            :param kwargs:
            :return:
            """
            # 开启会话
            session = cls.mysql_client()
            try:
                result = func(cls, *args, session=session, **kwargs)
                # 提交事务
                session.commit()
                return result
            except Exception as e:
                # 异常事务回滚
                session.rollback()
                raise e
            finally:
                # 关闭会话，收回连接
                session.close()
        return wrapper
    return decorator


def df_to_html(caption, df):
    """DataFrame 对象转化为 table"""
    caption = '<caption>{}</caption>'.format(caption)
    html = ['<table border=1 style="font-family:Arial">\n', caption, '<tr>']
    for column in df.columns:
        html.append('<th>' + column + '</th>')
    html.append('</tr>\n')
    for row in df.values:
        tr = ['<tr>']
        for col in row:
            tr.append('<td>' + str(col) + '</td>')
        tr.append('</tr>\n')
        html.append(''.join(tr))
    html.append('</table>\n')
    return ''.join(html)
